@@ -28,62 +28,50 @@
 (* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *)
 (**************************************************************************)
 
-open Lexing
-open Lexer
 open Misc
 
-(* If I could rewrite Capucine, I would change the following.
-   - Typing would use a union-find data structure, and each type would
-     actually be a key of this union-find. Keys would thus be abstract
-     and we would never forget an expand_head.
-   - The distinction between terms and predicates would not be done
-     by the parser, but by a later pass.
-   - The AST for terms would be different. Never be afraid to create more
-     ASTs.
-   - More thought for interp.ml, especially for how to read
-     region references and term/expr contexts. *)
+let handle_file file =
+  log "----- PARSING -----@.";
+  let past = Lang.file_of_file file in
+  Lang.print_file false Format.std_formatter past;
+  log "@.----- TYPING AND INFERENCE -----@.";
+  let time = Unix.gettimeofday () in
+  let env, tast = Infer.file past in
+  let delay = Unix.gettimeofday () -. time in
+  log "@.Time spent on inference: %F@." delay;
+  Pp.pp_inferred Format.std_formatter tast;
+  log "@.----- WHY INTERPRETATION -----@.";
+  let file = Interp.file env tast in
+  match !Options.output_file with
+    | None ->
+        Why.to_channel stdout file
+    | Some f ->
+        Why.to_file f file
 
-let parse file =
-  let ch = try
-    open_in file
-  with
-    | Sys_error sys ->
-        error "Cannot open file (%s)" sys
-  in
-  let lexbuf = Lexing.from_channel ch in
+let handle_error f x =
   try
-    lexbuf.lex_curr_p <- { lexbuf.lex_curr_p with pos_fname = file };
-    let ast = Parser.file Lexer.token lexbuf in
-    close_in ch;
-    ast
+    f x
   with
-    | Lex_error (loc, err)
-    | Misc.Parse_error (loc, err) ->
-        Loc.locate_error loc "%s" err
-    | Parsing.Parse_error ->
-        Loc.locate_error (lexbuf_location lexbuf) "Parse error"
-
-let file file =
-  log "========= File %s ==========\n%!" file;
-  log "--------- Parsing ----------\n%!";
-  let ast = parse file in
-  log "--------- Binding ----------\n%!";
-  let _, ast = Bind.file ast in
-  Pp.file (Format.formatter_of_out_channel stdout) ast;
-  log "--------- Typing ----------\n%!";
-  let env, ast = Typing.file ast in
-  Pp.file (Format.formatter_of_out_channel stdout) ast;
-  log "--------- Inference ----------\n%!";
-  Infer.classes_environment := env.Typing_common.classes;
-  let ast = Infer.file ast in
-  Pp.file (Format.formatter_of_out_channel stdout) ast;
-  log "--------- Why Interpretation ----------\n%!";
-  Interp.classes_environment := env.Typing_common.classes;
-  Interp.values_environment := env.Typing_common.values;
-  let why = Interp.file ast in
-  Pp.print_unique_identifiers := true;
-  Why.to_channel stdout why;
-  Opt.iter (fun s -> Why.to_file s why) !Options.out
+    | Lang_ast.Parse_error (loc, msg) ->
+        loge "%aParse error: %s@." Lang.print_location loc msg;
+        exit 1
+    | Types.Type_error (loc, msg) ->
+        loge "%a%s@." Lang.print_location loc msg;
+        exit 1
+    | Assert_failure (file, line, char) ->
+        loge "File \"%s\", line %d, characters %d-%d:@.Assert failure.@."
+          file line char (char+1);
+        exit 2
+    | Match_failure (file, line, char) ->
+        loge "File \"%s\", line %d, characters %d-%d:@.Match failure.@."
+          file line char (char+1);
+        exit 2
+(*    | exn ->
+        loge "Internal error, see error message below or above.@.";
+        raise exn*)
 
 let () =
-  List.iter file !Options.files
+  Options.init ();
+  match !Options.file with
+    | "" -> Options.usage ()
+    | file -> handle_error handle_file file
