@@ -33,7 +33,19 @@ open Lexing
 
 module StringSet = Set.Make(String)
 module StringMap = Map.Make(String)
-module IdentSet = Set.Make(Ident)
+
+module IMap = Map.Make(Ident)
+module ISet = Set.Make(Ident)
+
+let imap_memo f =
+  let tbl = ref IMap.empty in
+  fun id ->
+    try
+      IMap.find id !tbl
+    with Not_found ->
+      let r = f id in
+      tbl := IMap.add id r !tbl;
+      r
 
 let error s =
   let buf = Buffer.create 128 in
@@ -41,15 +53,31 @@ let error s =
   kfprintf
     (fun _ ->
        fprintf fmt "@?";
-       eprintf "%s\n" (Buffer.contents buf);
+       printf "@?";
+       eprintf "%s@." (Buffer.contents buf);
        exit 1)
     fmt
     s
 
-let lexbuf_location lexbuf =
-  lexeme_start_p lexbuf, lexeme_end_p lexbuf
+let warnings = ref []
 
-exception Parse_error of Loc.t * string
+let warning s =
+  let buf = Buffer.create 128 in
+  let fmt = formatter_of_buffer buf in
+  kfprintf
+    (fun _ ->
+      fprintf fmt "@?";
+      warnings := Buffer.contents buf :: !warnings)
+    fmt
+    s
+
+let print_warnings () =
+  List.iter
+    (printf "%s@.")
+    (List.rev !warnings)
+
+let () =
+  at_exit print_warnings
 
 let rec list_fold_map f acc env = function
   | [] -> env, List.rev acc
@@ -99,8 +127,14 @@ let rec list_mapi ?(i = 0) ?(acc = []) f = function
   | [] -> List.rev acc
   | x :: r -> list_mapi ~i: (i + 1) ~acc: (f i x :: acc) f r
 
-let check loc b x =
-  ksprintf (fun s -> if b then () else Loc.locate_error loc "%s" s) x
+let rec list_map3 ?(acc = []) f a b c =
+  match a, b, c with
+    | [], [], [] ->
+        List.rev acc
+    | xa :: ra, xb :: rb, xc :: rc ->
+        list_map3 ~acc: (f xa xb xc :: acc) f ra rb rc
+    | _ ->
+        invalid_arg "list_map3"
 
 let (|>) x f = f x
 
@@ -123,11 +157,18 @@ module Opt = struct
 
   let if_none v = function
     | None -> v
-    | Some _ as x -> x
+    | Some x -> x
 end
 
+(* option list flattening *)
+let olf = function
+  | None -> []
+  | Some l -> l
+
 let log s = kfprintf (fun fmt -> ()) std_formatter s
 
+let loge s = kfprintf (fun fmt -> ()) err_formatter s
+
 let fresh =
   let i = ref (-1) in
   fun () ->
@@ -143,3 +184,7 @@ let memoize f =
       let r = f x in
       Hashtbl.add tbl x r;
       r
+
+let dummy_loc name =
+  let p = { Lexing.dummy_pos with Lexing.pos_fname = name } in
+  p, p
