@@ -31,43 +31,98 @@
 open Ocamlbuild_plugin
 
 let native = false
+let menhir = false
+let hat_depth = max_int 
 
 let nativify x =
   if native then x ^ ".native" else x ^ ".byte"
 
-let enter_build_dir () =
-  Printf.printf
-    "Ocamlbuild: Entering directory `%s'\n"
-    (if Filename.is_relative !Options.build_dir then
-       Filename.concat (Sys.getcwd ()) !Options.build_dir
-     else
-       !Options.build_dir)
+let capucine_main = nativify "main"
+
+let capucine ?o ?(mode: [ `Why2 | `Why3 | `Boron ] = `Why2) file =
+  S [
+    begin
+      if not native then
+        S [ A "ocamlrun"; A "-bt" ]
+      else
+        N
+    end;
+    A ("./" ^ capucine_main);
+    A "-hat"; A (string_of_int hat_depth);
+    begin
+      match o with
+        | None -> N
+        | Some o -> S [ A "-o"; P o ]
+    end;
+    begin
+      match mode with
+        | `Why2 -> A "-why2"
+        | `Why3 -> A "-why3"
+        | `Boron -> A "-boron"
+    end;
+    P file;
+  ]
+
+(* --coq-preamble has no effect on my computer :( *)
+let coq_preamble = "Require Export Why.
+Require Export capucine_why."
 
 let () = dispatch begin function
+  | Before_options ->
+      if menhir then
+        begin
+          Options.use_menhir := true;
+          tag_any [ "explain" ]
+        end
   | After_rules ->
       flag [ "ocamlyacc" ] (A "-v");
 
-      rule "capucine: cap -> why"
+      rule "parsini: X.pini -> X_lexer.mll, X_parser.mly, X_ast.ml, X.ml, X.mli"
         ~deps: [
-          if native then nativify "main" else "main.byte";
-          "%.cap";
+          "%.pini";
         ]
+        ~prods: [
+          "%_lexer.mll";
+          "%_parser.mly";
+          "%_ast.ml";
+          "%.ml";
+          "%.mli";
+        ]
+        (fun env _ ->
+           let pini = env "%.pini" in
+           Cmd (S [ A "ocamlrun"; A "-bt";
+                    A "parsini"; A "-variant"; A "mono";
+                    P pini ]));
+
+      rule ".cat"
+        ~dep: "%"
+        ~prod: "%.cat"
+        (fun env _ ->
+          let f = env "%" in
+          Cmd (S [ A "cat"; P f ]));
+
+      rule ".test"
+        ~deps: [ capucine_main; "%.cap" ]
+        ~prod: "%.test"
+        (fun env _ ->
+          let cap = env "%.cap" in
+          Cmd (capucine cap));
+
+      (* la suite est copiee de capucine 1 *)
+
+      rule "capucine: cap -> why"
+        ~deps: [ capucine_main; "%.cap" ]
         ~prod: "%.why"
         (fun env _ ->
            let cap = env "%.cap" and why = env "%.why" in
-           Cmd (S [ A "ocamlrun"; A "-bt";
-                    A (nativify "./main"); A "-boron"; A "-o"; P why; P cap ]));
+           Cmd (capucine ~o: why ~mode: `Why2 cap));
 
       rule "capucine: cap -> mlw"
-        ~deps: [
-          if native then nativify "main" else "main.byte";
-          "%.cap";
-        ]
+        ~deps: [ capucine_main; "%.cap" ]
         ~prod: "%.mlw"
         (fun env _ ->
            let cap = env "%.cap" and why = env "%.mlw" in
-           Cmd (S [ A "ocamlrun"; A "-bt";
-                    A (nativify "./main"); A "-why3"; A "-o"; P why; P cap ]));
+           Cmd (capucine ~o: why ~mode: `Why3 cap));
 
       rule "gwhy: why -> gwhy"
         ~deps: [
@@ -78,14 +133,14 @@ let () = dispatch begin function
         (fun env _ ->
 (*           enter_build_dir ();*)
            let why = env "%.why" in
-           Cmd (S [ A "gwhy-bin"; (*A "--exp"; A "goal";*) A "--split-user-conj"; (*A "--all-vc";*) P "lib/capucine.why"; P why ]));
+           Cmd (S [ A "gwhy-bin"; (*A "--exp"; A "goal";*)
+                    A "--split-user-conj"; (*A "--all-vc";*)
+                    P "lib/capucine.why"; P why ]));
 
-(*
       rule "capucine.why -> capucine_why3.why"
         ~dep: "lib/capucine.why"
         ~prod: "lib/capucine_why3.why"
         (fun _ _ -> Cmd (S [ A "why"; A "-why3"; P "lib/capucine.why" ]));
-*)
 
       rule "why3: mlw -> prouveurs"
         ~deps: [
@@ -112,13 +167,40 @@ let () = dispatch begin function
 
       rule "coq: why -> coq"
         ~deps: [
-          "lib/capucine.why";
+          "lib/capucine_why.v";
           "%.why";
         ]
         ~prod: "%.coq"
         (fun env _ ->
 (*           enter_build_dir ();*)
            let why = env "%.why" in
-           Cmd (S [ A "why"; A "--coq"; (*A "--all-vc";*) P "lib/capucine.why"; P why ]));
+           Cmd (S [ A "why"; A "--coq"; (*A "--all-vc";*)
+                    P "lib/capucine.why";
+                    A "--coq-preamble"; A coq_preamble;
+                    P why ]));
+
+      rule "capucine.why -> capucine_why.v"
+        ~dep: "lib/capucine.why"
+        ~prod: "lib/capucine_why.v"
+        (fun env _ ->
+          Cmd (S [ A "why"; A "--coq"; P "lib/capucine.why" ]));
+
+      rule "v -> vo"
+        ~dep: "%.v"
+        ~prod: "%.vo"
+        (fun env _ ->
+          let file = env "%.v" in
+          Cmd (S [ A "coqc"; P file ]));
+
+      rule "coqide"
+        ~deps: [
+          "lib/capucine_why.vo";
+          "%.coq";
+        ]
+        ~prod: "%.coqide"
+        (fun env _ ->
+(*           enter_build_dir ();*)
+           let file = env "%_why.v" in
+           Cmd (S [ A "coqide"; A "-I"; P "lib"; P file ]))
   | _ -> ()
 end
