@@ -31,6 +31,8 @@
 let boron = ref false
 let use_boron b = boron := b
 
+module ISet = Set.Make(Ident)
+
 (******************************************************************************)
 (*                                    AST                                     *)
 (******************************************************************************)
@@ -143,7 +145,13 @@ let a_bool b = ABool b
 let a_true = a_bool true
 let a_false = a_bool false
 let a_app x args = APred (x, args)
-let a_and x y = AAnd (x, y)
+
+let a_and x y =
+  match x, y with
+    | ABool true, z
+    | z, ABool true -> z
+    | _ -> AAnd (x, y)
+
 let a_or x y = AOr (x, y)
 let a_iff x y = AIff (x, y)
 let a_implies x y = AImplies (x, y)
@@ -211,9 +219,10 @@ type expr =
   | EAssign of Ident.t * expr
   | ETuple of expr list
   | EProj of int * int * expr
+  | ERef of expr
 
 let rec may_have_side_effects = function
-  | EVoid | EInt _ | EReal _ | EBool _ | EVar _ | EDeref _ ->
+  | EVoid | EInt _ | EReal _ | EBool _ | EVar _ | EDeref _ | ERef _ ->
       false
   | EAnd (a, b)
   | EOr (a, b)
@@ -385,15 +394,88 @@ let term_proj n i t = LProj (n, i, t)
 let expr_tuple l = ETuple l
 let expr_proj n i e = EProj (n, i, e)
 
+let expr_ref e = ERef e
+
+let rec term_reads acc t =
+  match t with
+    | LVoid
+    | LInt _
+    | LReal _
+    | LBool _
+    | LVar _
+    | LResult
+      -> acc
+
+    | LProj (_, _, a)
+    | LNot a
+    | LNeg a
+      -> term_reads acc a
+
+    | LEq (a, b)
+    | LLet (_, a, b)
+    | LAnd (a, b)
+    | LOr (a, b)
+    | LXor (a, b)
+    | LAdd (a, b)
+    | LSub (a, b)
+    | LDiv (a, b)
+    | LMul (a, b)
+    | LMod (a, b)
+    | LLt (a, b)
+    | LGt (a, b)
+    | LLe (a, b)
+    | LGe (a, b)
+      -> term_reads (term_reads acc a) b
+
+    | LIf (a, b, c)
+      -> term_reads (term_reads (term_reads acc a) b) c
+
+    | LTuple l
+    | LApp (_, l)
+      -> List.fold_left term_reads acc l
+
+    | LDeref (x, _)
+      -> ISet.add x acc
+
+let rec assertion_reads acc a =
+  match a with
+    | ABool _
+      -> acc
+    | AAnd (a, b)
+    | AOr (a, b)
+    | AIff (a, b)
+    | AImplies (a, b)
+      -> assertion_reads (assertion_reads acc a) b
+    | AIf (t, a, b)
+      -> assertion_reads (assertion_reads (term_reads acc t) a) b
+    | ALet (_, t, a)
+      -> assertion_reads (term_reads acc t) a
+    | ANot a
+    | AForall (_, _, a)
+    | AExists (_, _, a)
+      -> assertion_reads acc a
+    | AEq (ta, tb)
+    | ANeq (ta, tb)
+    | AGt (ta, tb)
+    | ALt (ta, tb)
+    | AGe (ta, tb)
+    | ALe (ta, tb)
+      -> term_reads (term_reads acc ta) tb
+    | APred (n, tl)
+      -> List.fold_left term_reads acc tl
+
 let assume = function
   | ABool true ->
       const_void
   | a ->
-      black_box (type_annot ~post: a (type_base unit))
+      let reads = assertion_reads ISet.empty a in
+      let reads = ISet.elements reads in
+      black_box (type_annot ~reads ~post: a (type_base unit))
 
 type decl =
   | DLogicType of Ident.t list * Ident.t
   | DLogic of Ident.t * logic_type list * logic_type
+  | DLogicDef of Ident.t * (Ident.t * logic_type) list * logic_type * term
   | DParam of Ident.t * typ
   | DLet of Ident.t * expr
   | DPredicate of Ident.t * (Ident.t * logic_type) list * assertion option
@@ -401,6 +483,7 @@ type decl =
 
 let decl_logic_type ?(args = []) name = DLogicType (args, name)
 let decl_logic name ?(args = []) t = DLogic (name, args, t)
+let decl_logic_def name ?(args = []) ty t = DLogicDef (name, args, ty, t)
 let decl_parameter name t = DParam (name, t)
 let decl_let name e = DLet (name, e)
 let decl_predicate name args body = DPredicate (name, args, body)
@@ -414,9 +497,50 @@ let file l = l
 (*                              Pretty-Printing                               *)
 (******************************************************************************)
 
-open Pp
 open Format
 
+module Pp =
+struct
+  let nothing fmt () = ()
+  let comma_space fmt () = fprintf fmt ",@ "
+  let string s fmt () = fprintf fmt "%s" s
+  let space fmt () = fprintf fmt "@ "
+
+  let list ?(sep = nothing) ?(left = nothing) ?(right = nothing) f fmt l =
+    match l with
+      | [] -> ()
+      | [x] ->
+          left fmt ();
+          f fmt x;
+          right fmt ();
+      | x :: rem ->
+          left fmt ();
+          f fmt x;
+          List.iter
+            (fun x ->
+              sep fmt ();
+              f fmt x)
+            rem;
+          right fmt ()
+
+  let option ?(left = nothing) ?(right = nothing) f fmt = function
+    | None -> ()
+    | Some x ->
+        left fmt ();
+        f fmt x;
+        right fmt ()
+
+  let capitalized_identifier fmt i =
+  (*  if !print_unique_identifiers then
+      fprintf fmt "%s" (String.capitalize (Ident.uname i))
+    else*)
+      fprintf fmt "%s" (String.capitalize (Ident.name i))
+end
+
+open Pp
+
+(******************************************************************************)
+
 (* Identifiers should not be keywords. *)
 let is_keyword = function
   | "array" -> true
@@ -424,9 +548,9 @@ let is_keyword = function
 
 let identifier fmt x =
   if is_keyword (Ident.name x) then
-    fprintf fmt "_%a" Pp.identifier x
+    fprintf fmt "_%a" Ident.pp x
   else
-    Pp.identifier fmt x
+    Ident.pp fmt x
 
 let prime_identifier fmt id =
   fprintf fmt "'%a" identifier id
@@ -467,17 +591,17 @@ module Why2 = struct
     (fun a b -> term_app id [ a; b ]),
     (fun a b -> app (app (var id) a) b)
 
-  let fst, expr_fst =
+  let term_fst, expr_fst =
     let id = Ident.create "fst" in
     (fun a -> term_app id [ a ]),
     (fun a -> app (var id) a)
 
-  let snd, expr_snd =
+  let term_snd, expr_snd =
     let id = Ident.create "snd" in
     (fun a -> term_app id [ a ]),
     (fun a -> app (var id) a)
 
-  let proj = make_proj fst snd
+(*  let proj = make_proj fst snd*)
 
   let rec pp_logic_type fmt t =
     let o x = fprintf fmt x in
@@ -510,8 +634,10 @@ module Why2 = struct
             o "(%s)" s
       | LBool b ->
           o "%b" b
-      | LTuple l -> pp_term fmt (make_tuple term_void pair l)
-      | LProj (n, i, t) -> pp_term fmt (make_proj fst snd n i t)
+      | LTuple l ->
+          pp_term fmt (make_tuple term_void pair l)
+      | LProj (n, i, t) ->
+          pp_term fmt (make_proj term_fst term_snd n i t)
       | LApp (f, args) ->
           o "@[<hv 2>%a%a@]"
             identifier f
@@ -763,11 +889,14 @@ module Why2 = struct
       | EGe (x, y) ->
           o "(@[%a >=@ %a@])" pp_expr x pp_expr y
       | EEq (x, y) ->
-          o "(@[%a =@ %a@])" pp_expr x pp_expr y
-      | ETuple l -> 
+          o "(@[[{} bool {if result then %a =@ %a@ else not %a =@ %a}]@])"
+            pp_expr x pp_expr y pp_expr x pp_expr y
+      | ETuple l ->
           pp_expr fmt (make_tuple const_void expr_pair l)
       | EProj (n, i, e) ->
           pp_expr fmt (make_proj expr_fst expr_snd n i e)
+      | ERef e ->
+          o "@[ref@ %a@]" pp_expr e
 
   let pp_decl fmt d =
     let o x = fprintf fmt x in
@@ -785,6 +914,15 @@ module Why2 = struct
             identifier name
             (list ~sep: comma_space pp_logic_type) args
             pp_logic_type res
+      | DLogicDef (name, args, res, body) ->
+          o "@[<hov 2>function %a(%a): %a =@ %a@]@.@."
+            identifier name
+            (list ~sep: comma_space
+               (fun fmt (id, ty) ->
+                 fprintf fmt "%a: %a" identifier id pp_logic_type ty))
+            args
+            pp_logic_type res
+            pp_term body
       | DParam (name, t) ->
           o "@[<hov 2>parameter %a:@ %a@]@.@." identifier name pp_type t
       | DLet (name, e) ->
@@ -824,6 +962,8 @@ module Why3 = struct
   let rec pp_logic_type fmt t =
     let o x = fprintf fmt x in
     match t with
+      | LTTuple [x] ->
+          pp_logic_type fmt x
       | LTTuple l ->
           o "@[<hov 2>(%a)@]"
             (list ~sep: comma_space pp_logic_type) l
@@ -1112,6 +1252,8 @@ module Why3 = struct
           o "x";
           for j = i+1 to n-1 do o ",_" done;
           o ")@ = %a in x)@]" pp_expr e
+      | ERef e ->
+          o "(@[ref@ %a@])" pp_expr e
 
   let pp_decl fmt d =
     let o x = fprintf fmt x in
@@ -1131,6 +1273,14 @@ module Why3 = struct
             (list ~left: space ~sep: space
                (fun fmt t -> fprintf fmt "(%a)" pp_logic_type t)) args
             pp_logic_type res
+      | DLogicDef (name, args, res, body) ->
+          o "{ @[<hov 2>logic %a%a:@ %a =@ %a@] }@.@."
+            identifier name
+            (list ~left: space ~sep: space
+               (fun fmt (id, ty) ->
+                 fprintf fmt "(%a: %a)" identifier id pp_logic_type ty)) args
+            pp_logic_type res
+            pp_term body
       | DParam (name, t) ->
           o "@[<hov 2>parameter %a:@ %a@]@.@."
             identifier name pp_type t
@@ -1155,15 +1305,15 @@ module Why3 = struct
   let to_channel ch file =
     let fmt = formatter_of_out_channel ch in
     fprintf fmt "{ use import int.ComputerDivision@\n  \
-                   use import capucine_why3.Why2 }@.@.";
+                   use import capucine_why3.Capucine_ctx }@.@.";
     List.iter (pp_decl fmt) file
 end
 
-let to_channel =
+let to_channel x =
   match !Options.output_kind with
-    | Options.Why2 -> use_boron false; Why2.to_channel
-    | Options.Why3 -> Why3.to_channel
-    | Options.Boron -> use_boron true; Why2.to_channel
+    | Options.Why2 -> use_boron false; Why2.to_channel x
+    | Options.Why3 -> Why3.to_channel x
+    | Options.Boron -> use_boron true; Why2.to_channel x
 
 let to_file name file =
   let ch = open_out name in
