@@ -172,6 +172,7 @@ val assign: Ident.t -> expr -> expr
 val expr_ignore: expr -> expr -> expr
 val expr_tuple: expr list -> expr
 val expr_proj: int -> int -> expr -> expr (** [expr_proj count index expr] *)
+val expr_ref: expr -> expr
 
 val assume: assertion -> expr
 
@@ -179,6 +180,8 @@ type decl
 
 val decl_logic_type: ?args: Ident.t list -> Ident.t -> decl
 val decl_logic: Ident.t -> ?args: logic_type list -> logic_type -> decl
+val decl_logic_def: Ident.t -> ?args: (Ident.t * logic_type) list ->
+  logic_type -> term -> decl
 val decl_parameter: Ident.t -> typ -> decl
 val decl_let: Ident.t -> expr -> decl
 val decl_predicate: Ident.t -> (Ident.t * logic_type) list ->
