@@ -35,17 +35,33 @@ let region =
   let id = Ident.create "region" in
   fun x -> logic_type ~args: [ x ] id
 
+(*let region = logic_type (Ident.create "region")*)
+
 let pointer = logic_type (Ident.create "pointer")
 
-let either =
-  let id = Ident.create "either" in
-  fun x y -> logic_type ~args: [ x; y ] id
+let term_empty, expr_empty =
+  let id = Ident.create "empty" in
+  term_var id, var id
 
-let set, expr_set =
+let in_region =
+  let id = Ident.create "in_region" in
+  (fun p r -> a_app id [ p; r ])
+
+let term_get, expr_get =
+  let id = Ident.create "get" in
+  (fun r p -> term_app id [ r; p ]),
+  (fun r p -> app (app (var id) r) p)
+
+let term_set, expr_set =
   let id = Ident.create "set" in
   (fun r p x -> term_app id [ r; p; x ]),
   (fun r p x -> app (app (app (var id) r) p) x)
 
+(*
+let either =
+  let id = Ident.create "either" in
+  fun x y -> logic_type ~args: [ x; y ] id
+
 let get, expr_get =
   let id = Ident.create "get" in
   (fun r p -> term_app id [ r; p ]),
@@ -101,3 +117,4 @@ let in_expr: expr context = {
   set = expr_set;
   empty = expr_empty;
 }
+*)
